/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.controller;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openepics.names.rest.api.v1.INames;
import org.openepics.names.rest.beans.FieldName;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.response.Response;
import org.openepics.names.rest.beans.response.ResponseBoolean;
import org.openepics.names.rest.beans.response.ResponseBooleanList;
import org.openepics.names.rest.beans.response.ResponsePageNameElements;
import org.openepics.names.service.NamesService;
import org.openepics.names.util.ExceptionUtil;
import org.openepics.names.util.LogUtil;
import org.openepics.names.util.ServiceHttpStatusException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.RestController;

import com.google.common.collect.Lists;

/**
 * This part of REST API provides name data for Naming application.
 *
 * @author Lars Johansson
 */
@RestController
@EnableAutoConfiguration
public class NamesController implements INames {

    // note
    //     global exception handler available

    private static final Logger LOGGER = Logger.getLogger(NamesController.class.getName());

    private NamesService namesService;

    @Autowired
    public NamesController(
            NamesService namesService) {
        this.namesService = namesService;
    }

    @Override
    public ResponseEntity<List<NameElement>> createNames(List<NameElementCommand> nameElements) {
        // validate authority
        //     naming user & admin
        // validate
        // do

        try {
            namesService.validateNamesCreate(nameElements);
            return new ResponseEntity<>(namesService.createNames(nameElements), Response.HEADER_JSON, HttpStatus.CREATED);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public ResponsePageNameElements readNames(
            Boolean deleted, FieldName[] queryFields, String[] queryValues,
            FieldName orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        try {
            return namesService.readNames(
                    deleted, queryFields, queryValues,
                    orderBy, isAsc, page, pageSize);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    @Override
    public ResponsePageNameElements readNames(
            String name,
            FieldName orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        try {
            return namesService.readNames(name, orderBy, isAsc, page, pageSize);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    @Override
    public ResponsePageNameElements readNamesSystemStructure(
            String mnemonicpath,
            FieldName orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        try {
            return namesService.readNamesSystemStructure(mnemonicpath, orderBy, isAsc, page, pageSize);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    @Override
    public ResponsePageNameElements readNamesDeviceStructure(
            String mnemonicpath,
            FieldName orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        try {
            return namesService.readNamesDeviceStructure(mnemonicpath, orderBy, isAsc, page, pageSize);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    @Override
    public ResponsePageNameElements readNamesHistory(
            String uuid,
            FieldName orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        try {
            return namesService.readNamesHistory(uuid, orderBy, isAsc, page, pageSize);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public String equivalenceName(String name) {
        try {
            return namesService.equivalenceName(name);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<ResponseBoolean> existsName(String name) {
        try {
            return new ResponseEntity<>(new ResponseBoolean(namesService.existsName(name)), Response.HEADER_JSON, HttpStatus.OK);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()), Response.HEADER_JSON, HttpStatus.OK);
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED), Response.HEADER_JSON, HttpStatus.OK);
        }
    }

    @Override
    public ResponseEntity<ResponseBoolean> isLegacyName(String name) {
        try {
            return new ResponseEntity<>(new ResponseBoolean(namesService.isLegacyName(name)), Response.HEADER_JSON, HttpStatus.OK);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()), Response.HEADER_JSON, HttpStatus.OK);
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED), Response.HEADER_JSON, HttpStatus.OK);
        }
    }

    @Override
    public ResponseEntity<ResponseBoolean> isValidToCreateName(String name) {
        try {
            return new ResponseEntity<>(new ResponseBoolean(namesService.isValidToCreateName(name)), Response.HEADER_JSON, HttpStatus.OK);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()), Response.HEADER_JSON, HttpStatus.OK);
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED), Response.HEADER_JSON, HttpStatus.OK);
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public ResponseEntity<ResponseBooleanList> validateNamesCreate(List<NameElementCommand> nameElements) {
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (NameElementCommand nameElement : nameElements) {
            try {
                namesService.validateNamesCreate(nameElement);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceHttpStatusException e) {
                LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                LOGGER.log(Level.SEVERE, e.getMessage());
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.HEADER_JSON, HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateNamesUpdate(List<NameElementCommand> nameElements) {
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (NameElementCommand nameElement : nameElements) {
            try {
                namesService.validateNamesUpdate(nameElement);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceHttpStatusException e) {
                LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                LOGGER.log(Level.SEVERE, e.getMessage());
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.HEADER_JSON, HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateNamesDelete(List<NameElementCommand> nameElements) {
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (NameElementCommand nameElement : nameElements) {
            try {
                namesService.validateNamesDelete(nameElement);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceHttpStatusException e) {
                LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                LOGGER.log(Level.SEVERE, e.getMessage());
                LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
                if (response) {
                    response = false;
                    reason = ExceptionUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, ExceptionUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.HEADER_JSON, HttpStatus.OK);
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public List<NameElement> updateNames(List<NameElementCommand> nameElements) {
        // validate authority
        //     naming user & admin
        // validate
        // do

        try {
            namesService.validateNamesUpdate(nameElements);
            return namesService.updateNames(nameElements);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public List<NameElement> deleteNames(List<NameElementCommand> nameElements) {
        // validate authority
        //     naming user & admin
        // validate
        // do

        try {
          namesService.validateNamesDelete(nameElements);
          return namesService.deleteNames(nameElements);
        } catch (ServiceHttpStatusException e) {
            LogUtil.logServiceHttpStatusException(LOGGER, Level.SEVERE, e);
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, e.getMessage());
            LogUtil.logStackTraceElements(LOGGER, Level.SEVERE, e);
            throw e;
        }
    }

}