/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.openepics.names.docker.ITUtil.AuthorizationChoice;
import org.openepics.names.docker.ITUtil.EndpointChoice;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.response.ResponseBoolean;
import org.openepics.names.rest.beans.response.ResponseBooleanList;
import org.openepics.names.rest.beans.response.ResponsePageStructureElements;
import org.openepics.names.util.EnumUtil.StructureChoice;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Utility class to help (Docker) integration tests for Naming and PostgreSQL.
 *
 * @author Lars Johansson
 */
public class ITUtilStructureElement {

    /**
     * This class is not to be instantiated.
     */
    private ITUtilStructureElement() {
        throw new IllegalStateException("Utility class");
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that array of structure elements has expected length.
     *
     * @param actual array of structure elements
     * @param expectedLength expected length of array
     */
    static void assertContentLength(StructureElement[] actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.length);
    }

    /**
     * Assert that array of structure elements has length one and structure element is available.
     *
     * @param actual array of structure elements
     * @return structure element
     */
    static StructureElement assertContentLengthOne(StructureElement[] actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual[0]);
        return actual[0];
    }

    /**
     * Assert that response page with list of structure elements has expected length.
     *
     * @param actual response page with list of structure elements
     * @param expectedLength expected length of list
     */
    static void assertContentLength(ResponsePageStructureElements actual, int expectedLength) {
        assertNotNull(actual);
        assertEquals(expectedLength, actual.getList().size());
        assertEquals(expectedLength, actual.getListSize());
    }

    /**
     * Assert that response page with list of structure elements has length one and structure element is available.
     *
     * @param actual response page with list of structure elements
     * @return structure element
     */
    static StructureElement assertContentLengthOne(ResponsePageStructureElements actual) {
        assertContentLength(actual, 1);
        assertNotNull(actual.getList().get(0));
        return actual.getList().get(0);
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Assert that arrays are equal with same length and same content in each array position.
     *
     * @param actual actual array of structure elements
     * @param expected expected arbitrary number of structure elements
     */
    static void assertEqualsStructureElements(StructureElement[] actual, StructureElement... expected) {
        if (expected != null) {
            assertNotNull(actual);
            assertEquals(expected.length, actual.length);
            for (int i=0; i<expected.length; i++) {
                assertTrue(expected[i].equals(actual[i]));
            }
        } else {
            assertNull(actual);
        }
    }

    /**
     * Assert content of structure element.
     *
     * @param actual structure element
     * @param expectedType expected type
     * @param expectedUuid expected uuid
     * @param expectedParent expected parent
     * @param expectedName expected name
     * @param expectedMnemonic expected mnemonic
     * @param expectedMnemonicpath expected mnemonic path
     * @param expectedLevel expected level
     * @param expectedDescription expected description
     * @param expectedStatus expected status
     * @param expectedIsLatest expected latest
     * @param expectedIsDeleted expected deleted
     * @param expectedWho expected who
     * @param expectedComment expected comment
     */
    static void assertContent(StructureElement actual,
            Type expectedType, UUID expectedUuid, UUID expectedParent,
            String expectedName, String expectedMnemonic, String expectedMnemonicpath, Integer expectedLevel,
            String expectedDescription, Status expectedStatus, Boolean expectedIsLatest, Boolean expectedIsDeleted,
            String expectedWho, String expectedComment) {
        assertNotNull(actual);
        assertEquals(expectedType, actual.getType());
        assertEquals(expectedUuid, actual.getUuid());
        assertEquals(expectedParent, actual.getParent());
        assertEquals(expectedName, actual.getName());
        assertEquals(expectedMnemonic, actual.getMnemonic());
        assertEquals(expectedMnemonicpath, actual.getMnemonicpath());
        assertEquals(expectedLevel, actual.getLevel());
        assertEquals(expectedDescription, actual.getDescription());
        assertEquals(expectedStatus, actual.getStatus());
        assertEquals(expectedIsLatest, actual.isLatest());
        assertEquals(expectedIsDeleted, actual.isDeleted());
        assertNotNull(actual.getWhen());
        assertEquals(expectedWho, actual.getWho());
        assertEquals(expectedComment, actual.getComment());
    }

    /**
     * Assert content of structure element.
     *
     * @param actual structure element
     * @param expected structure element command
     */
    static void assertContent(StructureElement actual, StructureElementCommand expected) {
        assertNotNull(actual);
        if (expected != null) {
            assertEquals(expected.getUuid(), actual.getUuid());
            assertEquals(expected.getType(), actual.getType());
            assertEquals(expected.getParent(), actual.getParent());
            assertEquals(expected.getName(), actual.getName());
            assertEquals(expected.getMnemonic(), actual.getMnemonic());
            assertEquals(expected.getDescription(), actual.getDescription());
            assertEquals(expected.getComment(), actual.getComment());
        }
    }


    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to read structure elements and assert number of items in list.
     *
     * @param queryString query string
     * @param expectedLessThanOrEqual (if non-negative number) less than or equal to this number of items
     * @param expectedGreaterThanOrEqual (if non-negative number) greater than or equal to this number of items
     * @return number of items in list
     */
    public static Integer assertRead(String queryString, int expectedLessThanOrEqual, int expectedGreaterThanOrEqual) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;
            ResponsePageStructureElements responsePageStructureElements = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + queryString);
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            responsePageStructureElements = mapper.readValue(response[1], ResponsePageStructureElements.class);
            if (expectedLessThanOrEqual >= 0) {
                assertTrue(responsePageStructureElements.getListSize() <= expectedLessThanOrEqual);
            }
            if (expectedGreaterThanOrEqual >= 0) {
                assertTrue(responsePageStructureElements.getListSize() >= expectedGreaterThanOrEqual);
            }

            return responsePageStructureElements.getListSize();
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
        return null;
    }

    /**
     * Utility method to get name equivalence of a mnemonic or name.
     *
     * @param name name
     * @param expected expected response
     */
    public static void assertEquivalence(String name, String expected) {
        try {
            String[] response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/equivalence/" + name);
            ITUtil.assertResponseLength2CodeOKContent(response, expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if structure exists.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     */
    public static void assertExists(Type type, String mnemonicpath, Boolean expected) {
        assertExists(type, mnemonicpath, expected, null);
    }

    /**
     * Utility method to check if structure exists.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     * @param expectedMessageEmpty expected message empty
     */
    public static void assertExists(Type type, String mnemonicpath, Boolean expected, Boolean expectedMessageEmpty) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/exists/" + type.toString() + "/" + mnemonicpath);
            ITUtil.assertResponseLength2CodeOK(response);
            if (expectedMessageEmpty != null) {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected, expectedMessageEmpty);
            } else {
                ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
            }
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to check if structure is valid to create.
     * Method will not produce a message.
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @param expected expected response
     */
    public static void assertIsValidToCreate(Type type, String mnemonicpath, Boolean expected) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            response = ITUtil.doGetJson(ITUtil.HTTP_IP_PORT_NAMING_API_V1_STRUCTURES + "/isvalidtocreate/" + type.toString() + "/" + mnemonicpath);
            ITUtil.assertResponseLength2CodeOK(response);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBoolean.class), expected);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to (try to) validate a structure element from given json and assert expected response.
     * Method expected to NOT succeed to validate
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param json json
     * @param structureChoice structure choice
     * @param responseCode response code
     */
    public static void assertValidate(String json, StructureChoice structureChoice, int responseCode) {
        String validatePath = getValidatePath(structureChoice);

        try {
            String[] response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, validatePath, json));
            ITUtil.assertResponseLength2Code(response, responseCode);
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to validate and assert expected response.
     *
     * @param structureElement structure element
     * @param structureChoice structure choice
     * @param expected expected response
     */
    public static void assertValidate(StructureElementCommand structureElement, StructureChoice structureChoice, Boolean expected) {
        String validatePath = getValidatePath(structureChoice);

        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, validatePath, "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), expected);
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to return validate path for structure choice.
     *
     * @param structureChoice structure choice
     * @return validate path
     */
    private static String getValidatePath(StructureChoice structureChoice) {
        String validatePath = "";
        switch (structureChoice) {
            case CREATE:
                validatePath = "/validatecreate";
                break;
            case UPDATE:
                validatePath = "/validateupdate";
                break;
            case DELETE:
                validatePath = "/validatedelete";
                break;
            case APPROVE:
                validatePath = "/validateapprove";
                break;
            case CANCEL:
                validatePath = "/validatecancel";
                break;
            case REJECT:
                validatePath = "/validatereject";
                break;
            default:
                break;
        }
        return validatePath;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Utility method to (try to) create a structure element and assert result response code.
     * Method expected to NOT succeed to create a structure element
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param structureElement structure element
     * @param responseCode response code
     * @throws JsonProcessingException
     */
    public static void assertCreate(StructureElementCommand structureElement, int responseCode) throws JsonProcessingException {
        ObjectMapper mapper = new ObjectMapper();
        assertCreate("[" + mapper.writeValueAsString(structureElement) + "]", responseCode);
    }

    /**
     * Utility method to (try to) create a structure element from given json and assert result response code.
     * Method expected to NOT succeed to create a structure element
     * but instead assert response code (possibly also a non-empty message).
     *
     * @param json json
     * @param responseCode response code
     */
    public static void assertCreate(String json, int responseCode) {
        try {
            String[] response = ITUtil.runShellCommand(ITUtil.curlPostPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "", json));
            ITUtil.assertResponseLength2Code(response, responseCode);
            // ITUtil.assertMessageNotEmpty(mapper.readValue(response[1], Response.class));
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    /**
     * Utility method to create a structure element and assert result.
     *
     * @param structureElement structure element
     * @return created structure element
     */
    public static StructureElement assertCreate(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;
            UUID uuid = null;

            StructureElement[] createdStructureElements  = null;
            StructureElement   createdStructureElement   = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validatecreate", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPostPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_CREATED);
            createdStructureElements = mapper.readValue(response[1], StructureElement[].class);
            createdStructureElement = ITUtilStructureElement.assertContentLengthOne(createdStructureElements);
            assertNotNull(createdStructureElement.getUuid());

            uuid = createdStructureElement.getUuid();
            structureElement.setUuid(uuid);

            ITUtilStructureElement.assertContent(createdStructureElement, structureElement);
            assertEquals(Status.PENDING, createdStructureElement.getStatus());
            assertEquals(Boolean.FALSE, createdStructureElement.isLatest());
            assertEquals(Boolean.FALSE, createdStructureElement.isDeleted());
            assertNotNull(createdStructureElement.getWhen());

            return createdStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to update a structure element and assert result.
     *
     * @param structureElement structure element
     * @return updated structure element
     */
    public static StructureElement assertUpdate(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;
            UUID uuid = null;

            StructureElement[] updatedStructureElements  = null;
            StructureElement   updatedStructureElement   = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validateupdate", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPutPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            updatedStructureElements = mapper.readValue(response[1], StructureElement[].class);
            updatedStructureElement = ITUtilStructureElement.assertContentLengthOne(updatedStructureElements);
            assertNotNull(updatedStructureElement.getUuid());

            uuid = updatedStructureElement.getUuid();
            structureElement.setUuid(uuid);

            ITUtilStructureElement.assertContent(updatedStructureElement, structureElement);
            assertEquals(Status.PENDING, updatedStructureElement.getStatus());
            assertEquals(Boolean.FALSE, updatedStructureElement.isLatest());
            assertEquals(Boolean.FALSE, updatedStructureElement.isDeleted());
            assertNotNull(updatedStructureElement.getWhen());

            return updatedStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to delete a structure element and assert result.
     *
     * @param structureElement structure element
     * @return deleted structure element
     */
    public static StructureElement assertDelete(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;
            UUID uuid = null;

            StructureElement[] deletedStructureElements  = null;
            StructureElement   deletedStructureElement   = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validatedelete", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlDeletePathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            deletedStructureElements = mapper.readValue(response[1], StructureElement[].class);
            deletedStructureElement = ITUtilStructureElement.assertContentLengthOne(deletedStructureElements);
            assertNotNull(deletedStructureElement.getUuid());

            uuid = deletedStructureElement.getUuid();
            structureElement.setUuid(uuid);

            ITUtilStructureElement.assertContent(deletedStructureElement, structureElement);
            assertEquals(Status.PENDING, deletedStructureElement.getStatus());
            assertEquals(Boolean.FALSE, deletedStructureElement.isLatest());
            assertEquals(Boolean.TRUE, deletedStructureElement.isDeleted());
            assertNotNull(deletedStructureElement.getWhen());

            return deletedStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element
     */
    public static StructureElement assertApprove(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            StructureElement[] approvedStructureElements = null;
            StructureElement   approvedStructureElement  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validateapprove", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/approve", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            approvedStructureElements = mapper.readValue(response[1], StructureElement[].class);
            approvedStructureElement = ITUtilStructureElement.assertContentLengthOne(approvedStructureElements);

            ITUtilStructureElement.assertContent(approvedStructureElement, structureElement);
            assertEquals(Status.APPROVED, approvedStructureElement.getStatus());
            assertEquals(Boolean.TRUE, approvedStructureElement.isLatest());
            // not known if deleted
            assertNotNull(approvedStructureElement.getWhen());

            return approvedStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element
     */
    public static StructureElement assertCancel(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            StructureElement[] cancelledStructureElements = null;
            StructureElement   cancelledStructureElement  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validatecancel", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/cancel", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            cancelledStructureElements = mapper.readValue(response[1], StructureElement[].class);
            cancelledStructureElement = ITUtilStructureElement.assertContentLengthOne(cancelledStructureElements);

            ITUtilStructureElement.assertContent(cancelledStructureElement, structureElement);
            assertEquals(Status.CANCELLED, cancelledStructureElement.getStatus());
            assertEquals(Boolean.FALSE, cancelledStructureElement.isLatest());
            // not known if deleted
            assertNotNull(cancelledStructureElement.getWhen());

            return cancelledStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element
     */
    public static StructureElement assertReject(StructureElementCommand structureElement) {
        try {
            ObjectMapper mapper = new ObjectMapper();
            String[] response = null;

            StructureElement[] rejectedStructureElements = null;
            StructureElement   rejectedStructureElement  = null;

            response = ITUtil.runShellCommand(ITUtil.curlGetPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/validatereject", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            ITUtil.assertEqualsResponseBoolean(mapper.readValue(response[1], ResponseBooleanList.class), Boolean.TRUE);

            response = ITUtil.runShellCommand(ITUtil.curlPatchPathJson(AuthorizationChoice.NONE, EndpointChoice.STRUCTURES, "/reject", "[" + mapper.writeValueAsString(structureElement) + "]"));
            ITUtil.assertResponseLength2Code(response, HttpURLConnection.HTTP_OK);
            rejectedStructureElements = mapper.readValue(response[1], StructureElement[].class);
            rejectedStructureElement = ITUtilStructureElement.assertContentLengthOne(rejectedStructureElements);

            ITUtilStructureElement.assertContent(rejectedStructureElement, structureElement);
            assertEquals(Status.REJECTED, rejectedStructureElement.getStatus());
            assertEquals(Boolean.FALSE, rejectedStructureElement.isLatest());
            // not known if deleted
            assertNotNull(rejectedStructureElement.getWhen());

            return rejectedStructureElement;
        } catch (IOException e) {
            fail();
        } catch (InterruptedException e) {
            fail();
        } catch (Exception e) {
            fail();
        }

        return null;
    }

    /**
     * Utility method to create, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after create)
     */
    public static StructureElement assertCreateApprove(StructureElementCommand structureElement) {
        return assertApprove(assertCreate(structureElement));
    }

    /**
     * Utility method to create, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after create)
     */
    public static StructureElement assertCreateCancel(StructureElementCommand structureElement) {
        return assertCancel(assertCreate(structureElement));
    }

    /**
     * Utility method to create, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after create)
     */
    public static StructureElement assertCreateReject(StructureElementCommand structureElement) {
        return assertReject(assertCreate(structureElement));
    }

    /**
     * Utility method to update, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after update)
     */
    public static StructureElement assertUpdateApprove(StructureElementCommand structureElement) {
        return assertApprove(assertUpdate(structureElement));
    }

    /**
     * Utility method to update, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after update)
     */
    public static StructureElement assertUpdateCancel(StructureElementCommand structureElement) {
        return assertCancel(assertUpdate(structureElement));
    }

    /**
     * Utility method to update, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after update)
     */
    public static StructureElement assertUpdateReject(StructureElementCommand structureElement) {
        return assertReject(assertUpdate(structureElement));
    }

    /**
     * Utility method to delete, approve a structure element and assert result.
     *
     * @param structureElement structure element
     * @return approved structure element (after delete)
     */
    public static StructureElement assertDeleteApprove(StructureElementCommand structureElement) {
        return assertApprove(assertDelete(structureElement));
    }

    /**
     * Utility method to delete, cancel a structure element and assert result.
     *
     * @param structureElement structure element
     * @return cancelled structure element (after delete)
     */
    public static StructureElement assertDeleteCancel(StructureElementCommand structureElement) {
        return assertCancel(assertDelete(structureElement));
    }

    /**
     * Utility method to delete, reject a structure element and assert result.
     *
     * @param structureElement structure element
     * @return rejected structure element (after delete)
     */
    public static StructureElement assertDeleteReject(StructureElementCommand structureElement) {
        return assertReject(assertDelete(structureElement));
    }

}
